@@ -1686,7 +1686,7 @@ uint64_t export_results_from_groebner_qq(
         evi[i-1]    =   i;
       }
     } else {
-      for (len_t i = 1; i <= ebl; ++i) {
+      for (len_t i = 1; i < ebl; ++i) {
         evi[i-1]    =   i;
       }
       for (len_t i = ebl+1; i < evl; ++i) {
@@ -1700,7 +1700,7 @@ uint64_t export_results_from_groebner_qq(
         len_t idx = gb->lmps[p];
         hm  = gb->hm[idx]+OFFSET;
         int32_t l = gb->modpolys[p]->len;
-        for(int32_t n = 0; n < nve; n++){
+        for(int32_t n = 0; n < nv; n++){
             exp[term * nve + n + elim_block_len] = gb->ldm[p * nv + n];
         }
         mpz_set(cf_qq[term], gb->modpolys[p]->lm);
