@@ -586,6 +586,10 @@ static void symbolic_preprocessing(
     while (mat->sz <= nrr + oesld) {
         mat->sz *=  2;
         mat->rr =   realloc(mat->rr, (unsigned long)mat->sz * sizeof(hm_t *));
+        if (mat->rr == NULL) {
+            fprintf(stderr, "Allocating memory for matrix failed,\n");
+            fprintf(stderr, "segmentation fault will follow.\n");
+        }
     }
     for (; i < oesld; ++i) {
         if (!sht->hd[i].idx) {
@@ -598,6 +602,10 @@ static void symbolic_preprocessing(
         if (mat->sz == nrr) {
             mat->sz *=  2;
             mat->rr  =  realloc(mat->rr, (unsigned long)mat->sz * sizeof(hm_t *));
+            if (mat->rr == NULL) {
+                fprintf(stderr, "Allocating memory for matrix failed,\n");
+                fprintf(stderr, "segmentation fault will follow.\n");
+            }
         }
         sht->hd[i].idx = 1;
         mat->nc++;
@@ -612,8 +620,8 @@ static void symbolic_preprocessing(
     mat->rbal =   mat->nrl;
 
     /* initialize memory for reducer bit arrays for tracing information */
-    mat->rba  = (rba_t **)malloc((unsigned long)mat->rbal * sizeof(rba_t *));
-    const unsigned long len = nrr / 32 + ((nrr % 32) != 0);
+    mat->rba  = (rba_t **)malloc((uint64_t)mat->rbal * sizeof(rba_t *));
+    const uint64_t len = nrr / 32 + ((nrr % 32) != 0);
     for (i = 0; i < mat->nrl; ++i) {
         mat->rba[i] = (rba_t *)calloc(len, sizeof(rba_t));
     }
@@ -769,5 +777,9 @@ static int preprocessing(
     } else {
         generate_matrix_from_trace(mat, bs, md);
     }
-    return 0;
+    if (mat->nc == 0) {
+        return 1;
+    } else {
+        return 0;
+    }
 }
